import logging
from typing import List, Dict, Type, Tuple
import tensorflow as tf
from tensorflow.keras.layers import Layer, MultiHeadAttention

from keras.layers.pooling import Pooling2D, Pooling1D, Pooling3D
from keras.saving.saved_model.load import RevivedLayer
from keras.layers.convolutional import Conv
from keras_data_format_converter.layers.confighandlers.dataformat import handle_data_format


onnx_channel_first_cant_run_on_cpu_layers = (Conv, Pooling1D, Pooling2D, Pooling3D)
layers_to_build_on_call = onnx_channel_first_cant_run_on_cpu_layers + (MultiHeadAttention,)


def convert_layer(current_layer: Layer, input_shape: List[int],
                  custom_layers: Dict[str, Type[tf.keras.layers.Layer]]) \
        -> Tuple[Layer, bool]:
    logger = logging.getLogger(__name__)
    config = current_layer.get_config()

    if isinstance(current_layer, RevivedLayer):
        logger.debug(f"Layer skipped, name: {current_layer.name}, type: {current_layer.__class__.__name__},"
                     f" input_shape: {current_layer.input_shape}, output_shape: {current_layer.output_shape}")
        is_built = True
        return current_layer, is_built

    if isinstance(current_layer, onnx_channel_first_cant_run_on_cpu_layers):
        config = handle_data_format('channels_last', config)

    converted_layer = type(current_layer).from_config(config)
    is_built = True
    if isinstance(current_layer, layers_to_build_on_call) or isinstance(current_layer, tuple(custom_layers.values())):
        is_built = False
        return converted_layer, is_built

    weights = current_layer.get_weights()
    converted_layer.build(input_shape)
    converted_layer.set_weights(weights)
    logger.debug(f"Layer created, name: {converted_layer.name}, type: {current_layer.__class__.__name__},"
                 f" input_shape: {current_layer.input_shape}, output_shape: {current_layer.output_shape}")
    return converted_layer, is_built



from typing import Any, Dict


def handle_data_format(target_data_format: str, config: Dict[str, Any]) -> Dict[str, Any]:
    if "data_format" in config:
        config["data_format"] = target_data_format
    return config




import logging


def configure_logger(verbose):
    logging_level = logging.INFO
    if verbose:
        logging_level = logging.DEBUG
    logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s %(message)s', level=logging_level,
                        datefmt='%Y-%m-%d %H:%M:%S')


import logging
from typing import List, Type, Dict, Optional

import tensorflow as tf
from keras import Model
from keras.layers import Permute
from tensorflow import keras

from keras_data_format_converter.modelconverter import ModelConverter, calculate_permute_values
from keras_data_format_converter.utils import configure_logger


def convert_channels_first_to_last(model: keras.Model,
                                   should_transform_inputs_and_outputs=False,
                                   custom_layers: Optional[Dict[str, Type[tf.keras.layers.Layer]]] = None,
                                   verbose: bool = False) \
        -> keras.Model:
    """
    Convert keras models from channels first to last

    :param custom_layers: dictionary of custom layers
    :type custom_layers: Optional[Dict[str, Type[tf.keras.layers.Layer]]]
    :param verbose: by default true, set to False to lower the logging level
    :type verbose: bool
    :param model: keras model to convert in channels first format
    :type model: tensorflow.keras.Model
    :return: converted keras model in channels last format
    :rtype: tensorflow.keras.Model
    """
    converted_model = _convert_channels(model, custom_layers, verbose)
    if should_transform_inputs_and_outputs:
        converted_model = _transform_inputs_and_outputs(converted_model)

    return converted_model


def _convert_channels(model: tf.keras.Model,
                      custom_layers: Optional[Dict[str, Type[tf.keras.layers.Layer]]], verbose: bool) \
        -> tf.keras.Model:
    # configure logger
    configure_logger(verbose)
    logger = logging.getLogger(__name__)

    if custom_layers is None:
        custom_layers = {}

    model_converter = ModelConverter(model, custom_layers)
    converted_model = model_converter.convert_model()
    return converted_model


def _transform_inputs_and_outputs(k_model: Model) -> Model:
    new_inputs = []
    new_old_inputs = []
    for k_input in k_model.inputs:
        new_input_shape = Permute(
            calculate_permute_values(k_input.shape, to_channel_first=False))(k_input).shape[1:]
        if k_input.shape[0] is None:
            new_input_tensor = tf.keras.Input(new_input_shape, name=k_input.name)
        else:
            new_input_tensor = tf.keras.Input(new_input_shape, name=k_input.name, batch_size=k_input.shape[0])
        back_to_channel_first_perm_values = calculate_permute_values(new_input_tensor.shape, to_channel_first=True)
        permute_after = Permute(back_to_channel_first_perm_values)(new_input_tensor)
        new_inputs.append(new_input_tensor)
        new_old_inputs.append(permute_after)

    new_outputs = []
    new_old_outputs = k_model.call(new_old_inputs)
    if not isinstance(new_old_outputs, list):
        new_old_outputs = [new_old_outputs]
    for output in new_old_outputs:
        perm_values = calculate_permute_values(output.shape, to_channel_first=False)
        new_outputs.append(Permute(perm_values)(output))

    return Model(inputs=new_inputs, outputs=new_outputs)



import logging
from typing import List, Dict, Union, Tuple, Type

import tensorflow as tf
from keras.engine.keras_tensor import KerasTensor
from keras.layers.core import TFOpLambda

from tensorflow import keras, TensorShape
from tensorflow.keras.layers import Layer, InputLayer, Input, MultiHeadAttention
from tensorflow.python.keras.engine.node import Node

from keras_data_format_converter.layers.layer_utils import convert_layer, onnx_channel_first_cant_run_on_cpu_layers


def calculate_permute_values(input_shape: TensorShape, to_channel_first: bool) -> List[int]:
    n_dims = len(input_shape)
    if to_channel_first:
        return [n_dims - 1] + list(range(1, n_dims - 1))
    else:
        return list(range(2, n_dims)) + [1]


class ModelConverter:
    """The ModelConverter object converts keras models to the requested data_format

    The conversion process is handling special layers with data_format specific values like Concatenate,
    BatchNormalization, Reshape, etc. The logic of handling those layers is as follows:
    every special layer after an input from inputs_to_transpose until a layer with shape rank of 2 is being handled.

    :param model: The model to convert
    :type model: tensorflow.keras.Model
    """

    def __init__(self, model: keras.Model, custom_layers: Dict[str, Type[tf.keras.layers.Layer]]):
        self.model: keras.Model = model

        # cache properties
        self._clear_cache()

        # logger
        self._logger = logging.getLogger(__name__)

        self.custom_layers = custom_layers

    def convert_model(self) -> keras.Model:
        """
        Converting model to the requested target_data_format

        :return: Converted model
        :rtype: tensorflow.keras.Model
        """

        self._clear_cache()
        output_tensors: List[tf.Tensor] = []
        for out_tensor in self.model.outputs:
            converted_tensor = self._convert_tensor(out_tensor)
            output_tensors.append(converted_tensor)
        converted_model = tf.keras.Model(inputs=self._model_input_tensors, outputs=output_tensors)
        return converted_model

    def _find_tensor_index_by_name(self, node_input_tensors: List[KerasTensor], tensor_name: str) -> int:
        node_input_tensor_names = []
        for tensor in node_input_tensors:
            if tensor.name in self._tensor_name_to_original_tensor_name:
                node_input_tensor_names.append(self._tensor_name_to_original_tensor_name[tensor.name])
            else:
                node_input_tensor_names.append(tensor.name)

        return node_input_tensor_names.index(tensor_name)

    def _fix_args(self, call_args: Tuple, node_input_tensors: List[KerasTensor]) -> Tuple[List, List[KerasTensor]]:
        new_tfoplambda_args = list(call_args)
        for i, arg in enumerate(call_args):
            if isinstance(arg, dict):
                new_arg_dict, node_input_tensors = self._fix_kwargs(arg, node_input_tensors)
                new_tfoplambda_args[i] = new_arg_dict
            elif isinstance(arg, KerasTensor):
                try:
                    tensor_index = self._find_tensor_index_by_name(node_input_tensors, arg.name)
                    new_tfoplambda_args[i] = node_input_tensors.pop(tensor_index)
                except (IndexError, ValueError):
                    pass
        return new_tfoplambda_args, node_input_tensors

    def _fix_kwargs(self, call_kwargs, node_input_tensors):
        new_tfoplambda_kwargs = call_kwargs.copy()
        for key, arg in call_kwargs.items():
            if isinstance(arg, KerasTensor):
                try:
                    tensor_index = self._find_tensor_index_by_name(node_input_tensors, arg.name)
                    new_tfoplambda_kwargs[key] = node_input_tensors.pop(tensor_index)
                except (IndexError, ValueError):
                    pass
        return new_tfoplambda_kwargs, node_input_tensors

    def _default_call_layer_by_args(self,
                                    converted_layer: Layer, call_args: Tuple, call_kwargs: Dict,
                                    node_input_tensors: List[KerasTensor]) -> KerasTensor:
        new_tfoplambda_args, input_tensors = self._fix_args(call_args, node_input_tensors)
        new_tfoplambda_kwargs, input_tensors = self._fix_kwargs(call_kwargs, node_input_tensors)
        assert len(input_tensors) == 0

        return converted_layer(*new_tfoplambda_args, **new_tfoplambda_kwargs)

    def _call_layer_by_args(
            self, converted_layer: Layer, node_input_tensors: Union[KerasTensor, List[KerasTensor]],
            current_layer: Layer) -> KerasTensor:

        call_args = current_layer.inbound_nodes[0].call_args
        call_kwargs = current_layer.inbound_nodes[0].call_kwargs
        if isinstance(current_layer, TFOpLambda) and current_layer.symbol == 'stack':
            return converted_layer(node_input_tensors, *current_layer.inbound_nodes[0].call_args[2:],
                                   **current_layer.inbound_nodes[0].call_kwargs)

        if not isinstance(node_input_tensors, list):
            node_input_tensors = [node_input_tensors]

        if isinstance(current_layer, TFOpLambda) and current_layer.symbol == 'concat':
            # default call expects [list_of_values] while for concat it is [[list_of_values]]
            call_args = [self._fix_args(call_args[0], node_input_tensors)[0]]

        return self._default_call_layer_by_args(converted_layer, call_args, call_kwargs, node_input_tensors)

    def _convert_tensor(self, tensor: tf.Tensor) -> tf.Tensor:
        tensor_id = id(tensor)
        converted_tensor = self._tensor_cache.get(tensor_id)
        if converted_tensor is not None:
            return converted_tensor

        current_node = self._get_node_from_tensor(tensor)
        current_layer = self._get_layer_from_tensor(tensor)
        # Creating first model input layer
        if isinstance(current_layer, InputLayer):
            input_tensor = self._convert_input_tensor(tensor)
            self._tensor_cache[tensor_id] = input_tensor
            self._model_input_tensors.append(input_tensor)
            return input_tensor

        # get all input tensors
        node_input_tensors = []
        parent_nodes = current_node.parent_nodes
        for parent_node in parent_nodes:
            output_tensor = parent_node.outputs
            node_input_tensor = self._convert_tensor(output_tensor)
            node_input_tensors.append(node_input_tensor)

        layer_input_shape = [node_input_tensor.shape for node_input_tensor in node_input_tensors]

        if len(node_input_tensors) == 1:
            node_input_tensors = node_input_tensors[0]
            layer_input_shape = layer_input_shape[0]

        layer_name = current_layer.name
        converted_layer = self._layer_cache.get(layer_name)
        converted_layer_built = True
        if converted_layer is None:
            # special case of build layers:
            converted_layer, converted_layer_built = convert_layer(current_layer, layer_input_shape, self.custom_layers)
            self._layer_cache[layer_name] = converted_layer

        if converted_layer_built:
            if isinstance(converted_layer, TFOpLambda):
                converted_tensor = self._call_layer_by_args(converted_layer, node_input_tensors, current_layer)
            else:
                converted_tensor = converted_layer(node_input_tensors)
        elif isinstance(current_layer, MultiHeadAttention):
            weights = current_layer.get_weights()
            converted_tensor = converted_layer(*node_input_tensors)
            converted_layer.set_weights(weights)
        elif isinstance(converted_layer, onnx_channel_first_cant_run_on_cpu_layers) and \
                current_layer.get_config()['data_format'] == 'channels_first':
            perm_values = calculate_permute_values(node_input_tensors.shape, to_channel_first=False)
            permute_before = tf.keras.layers.Permute(perm_values)(node_input_tensors)
            tensor = converted_layer(permute_before)
            weights = current_layer.get_weights()
            converted_layer.set_weights(weights)
            perm_values = calculate_permute_values(tensor.shape, to_channel_first=True)
            permute_after = tf.keras.layers.Permute(perm_values)(tensor)
            self._tensor_name_to_original_tensor_name[permute_after.name] = tensor.name
            converted_tensor = permute_after
        elif type(converted_layer) in self.custom_layers.values():
            converted_tensor = self._call_layer_by_args(converted_layer, node_input_tensors, current_layer)
            weights = current_layer.get_weights()
            converted_layer.set_weights(weights)
        else:
            weights = current_layer.get_weights()
            converted_tensor = converted_layer(node_input_tensors)
            converted_layer.set_weights(weights)

        self._tensor_cache[tensor_id] = converted_tensor

        return converted_tensor

    @staticmethod
    def _get_layer_from_tensor(tensor: tf.Tensor) -> Layer:
        history = tensor._keras_history
        layer = history.layer
        return layer

    @staticmethod
    def _get_node_from_tensor(tensor: tf.Tensor) -> Node:
        history = tensor._keras_history
        layer = history.layer
        node_index = history.node_index
        node = layer.inbound_nodes[node_index]
        return node

    def _convert_input_tensor(self, input_tensor: tf.Tensor) -> tf.Tensor:
        input_layer = input_tensor._keras_history.layer
        inp_config = input_layer.get_config()
        if 'name' in inp_config and '.' in inp_config['name']:
            new_name_without_dots = inp_config['name'].replace(".", "_")
            self._tensor_name_to_original_tensor_name[new_name_without_dots] = inp_config['name']
            inp_config['name'] = new_name_without_dots
        new_input_tensor = Input(**inp_config)
        self._logger.debug(f"Input created, name: {new_input_tensor.name}, shape: {new_input_tensor.shape}")
        return new_input_tensor

    def _clear_cache(self) -> None:
        self._layer_cache: Dict[str, Layer] = {}
        self._tensor_cache: Dict[int, tf.Tensor] = {}
        self._model_input_tensors: List[tf.Tensor] = []
        self._tensor_name_to_original_tensor_name: Dict[str, str] = {}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['keras_data_format_converter',
 'keras_data_format_converter.layers',
 'keras_data_format_converter.layers.confighandlers']

package_data = \
{'': ['*']}

install_requires = \
['libclang==14.0.1', 'protobuf==3.20.1']

extras_require = \
{':platform_machine == "arm64"': ['tensorflow-macos==2.8.0'],
 ':platform_machine == "x86_64"': ['tensorflow==2.8.0',
                                   'tensorflow-addons>=0.17.1,<0.18.0']}

setup_kwargs = {
    'name': 'keras-data-format-converter',
    'version': '0.1.5',
    'description': 'Generates equal keras models with the desired data format',
    'long_description': '# Keras data format converter\n\nGenerates equal keras models with the desired data format  \n\n\n## Requirements\ntensorflow >= 2.0\n\n\n## API\n`convert_channels_first_to_last(model: keras.Model, inputs_to_transpose: List[str] = None, verbose: bool = False) -> keras.Model`\n\n`convert_channels_last_to_first(model: tf.keras.Model, inputs_to_transpose: List[str] = None, verbose: bool = False) \\\n        -> tf.keras.Model`\n\n`model`: Keras model to convert\n\n`inputs_to_transpose`: list of input names that need to be transposed due tothe data foramt changing  \n\n`verbose`: detailed output\n\n## Getting started\n\n```python\nfrom tensorflow import keras\nfrom keras_data_format_converter import convert_channels_last_to_first\n\n# Load Keras model\nkeras_model = keras.models.load_model("my_image_model")\n\n# Call the converter (image_input is an input that needs to be transposed, can be different for your model)\nconverted_model = convert_channels_last_to_first(keras_model, ["image_input"])\n```\n\n## Supported Layers with Special handling\n- [X] Normalization layers\n- [x] Permute\n- [x] Reshape\n- [x] Concatenate\n- [ ] Dot\n- [ ] MultiHeadAttention\n- [ ] TFOpLambda (Inserted by the Functional API construction whenever users call\n  a supported TF symbol on KerasTensors, see [here](https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/keras/layers/core.py#L1284) at Tensorflow repo for more info)\n\n## Unsupported Layers due to lack of data_format property\n- Cropping1D\n- Upsampling1D\n- Zeropadding1D\n- All layers in tensorflow.keras.preprocessing\n\n## How to deploy\n- Create a new release version on GitHub\n- Update parameters in setup.py (usually `version` and `download_url`)\n- Run `python setup.py sdist` in root directory\n- Run `pip install twine`\n- Run `twine upload dist/*`\n \n\n\n## License\nThis software is covered by MIT License.\n',
    'author': 'dorhar',
    'author_email': 'doron.harnoy@tensorleap.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
